@@ -18,6 +18,7 @@ withFdCache _ f = f Nothing
 module Network.Wai.Handler.Warp.FdCache (
     withFdCache
   , getFd
+  , getFd'
   , MutableFdCache
   , Refresh
   ) where
@@ -69,10 +70,10 @@ type Hash = Int
 type FdCache = MMap Hash FdEntry
 
 -- | Mutable Fd cacher.
-type MutableFdCache = Reaper FdCache (Hash, FdEntry)
+newtype MutableFdCache = MutableFdCache (Reaper FdCache (Hash, FdEntry))
 
 fdCache :: MutableFdCache -> IO FdCache
-fdCache mfc = reaperRead mfc
+fdCache (MutableFdCache reaper) = reaperRead reaper
 
 look :: MutableFdCache -> FilePath -> Hash -> IO (Maybe FdEntry)
 look mfc path key = searchWith key check <$> fdCache mfc
@@ -98,7 +99,7 @@ withFdCache duration action = bracket (initialize duration)
 -- The first argument is a cache duration in second.
 initialize :: Int -> IO (Maybe MutableFdCache)
 initialize 0 = return Nothing
-initialize duration = Just <$> mkReaper defaultReaperSettings
+initialize duration = Just . MutableFdCache <$> mkReaper defaultReaperSettings
     { reaperAction = clean
     , reaperDelay = duration
     , reaperCons = uncurry insert
@@ -131,24 +132,26 @@ prune k (Tom ent@(FdEntry _ fd mst) vs) = status mst >>= prune'
 
 terminate :: Maybe MutableFdCache -> IO ()
 terminate Nothing = return ()
-terminate (Just mfc) = do
-    !t <- reaperStop mfc
+terminate (Just (MutableFdCache reaper)) = do
+    !t <- reaperStop reaper
     mapM_ closeIt $ toList t
   where
     closeIt (_, FdEntry _ fd _) = closeFd fd
 
 ----------------------------------------------------------------
 
--- | Getting 'Fd' and 'Refresh' from the mutable Fd cacher.
 getFd :: MutableFdCache -> FilePath -> IO (Fd, Refresh)
-getFd mfc path = look mfc path key >>= getFd'
+getFd mfc path = getFd' mfc (hash path) path
+
+-- | Getting 'Fd' and 'Refresh' from the mutable Fd cacher.
+getFd' :: MutableFdCache -> Hash -> FilePath -> IO (Fd, Refresh)
+getFd' mfc@(MutableFdCache reaper) h path = look mfc path h >>= get
   where
-    key = hash path
-    getFd' Nothing = do
+    get Nothing = do
         ent@(FdEntry _ fd mst) <- newFdEntry path
-        reaperAdd mfc (key, ent)
+        reaperAdd reaper (h, ent)
         return (fd, refresh mst)
-    getFd' (Just (FdEntry _ fd mst)) = do
+    get (Just (FdEntry _ fd mst)) = do
         refresh mst
         return (fd, refresh mst)
 #endif
