@@ -9,6 +9,7 @@ module Network.Wai.Handler.Warp.HTTP2.Receiver (frameReceiver) where
 #if __GLASGOW_HASKELL__ < 709
 import Control.Applicative
 #endif
+import Control.Concurrent
 import Control.Concurrent.STM
 import qualified Control.Exception as E
 import Control.Monad (when, unless, void)
@@ -16,7 +17,7 @@ import Data.ByteString (ByteString)
 import qualified Data.ByteString as BS
 import Data.Maybe (isJust)
 import Network.HTTP2
-import Network.HTTP2.Priority
+import Network.HTTP2.Priority (toPrecedence, delete, prepare)
 import Network.Wai.Handler.Warp.HTTP2.EncodeFrame
 import Network.Wai.Handler.Warp.HTTP2.HPACK
 import Network.Wai.Handler.Warp.HTTP2.Request
@@ -27,7 +28,7 @@ import Network.Wai.Handler.Warp.Types
 ----------------------------------------------------------------
 
 frameReceiver :: Context -> MkReq -> (BufSize -> IO ByteString) -> IO ()
-frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
+frameReceiver ctx mkreq recvN = loop 0 `E.catch` sendGoaway
   where
     Context{ http2settings
            , streamTable
@@ -35,28 +36,33 @@ frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
            , continued
            , currentStreamId
            , inputQ
-           , outputQ
+           , controlQ
            } = ctx
     sendGoaway e
       | Just (ConnectionError err msg) <- E.fromException e = do
           csid <- readIORef currentStreamId
-          let frame = goawayFrame csid err msg
-          enqueueControl outputQ 0 $ OGoaway frame
+          let !frame = goawayFrame csid err msg
+          enqueueControl controlQ $ CGoaway frame
       | otherwise = return ()
 
     sendReset err sid = do
-        let frame = resetFrame err sid
-        enqueueControl outputQ 0 $ OFrame frame
+        let !frame = resetFrame err sid
+        enqueueControl controlQ $ CFrame frame
 
-    loop = do
+    loop :: Int -> IO ()
+    loop !n
+      | n == 6 = do
+          yield
+          loop 0
+      | otherwise = do
         hd <- recvN frameHeaderLength
         if BS.null hd then
-            enqueueControl outputQ 0 OFinish
+            enqueueControl controlQ CFinish
           else do
             cont <- processStreamGuardingError $ decodeFrameHeader hd
-            when cont loop
+            when cont $ loop (n + 1)
 
-    processStreamGuardingError (FrameHeaders, FrameHeader{streamId})
+    processStreamGuardingError (_, FrameHeader{streamId})
       | isResponse streamId = E.throwIO $ ConnectionError ProtocolError "stream id should be odd"
     processStreamGuardingError (FrameUnknown _, FrameHeader{payloadLength}) = do
         mx <- readIORef continued
@@ -92,7 +98,7 @@ frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
           control ftyp header pl ctx
       | otherwise = do
           checkContinued
-          strm@Stream{streamState,streamContentLength,streamPrecedence} <- getStream
+          !strm@Stream{streamState,streamContentLength,streamPrecedence} <- getStream
           pl <- recvN payloadLength
           state <- readIORef streamState
           state' <- stream ftyp header pl ctx state strm
@@ -105,7 +111,7 @@ frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
                               E.throwIO $ StreamError ProtocolError streamId
                           writeIORef streamPrecedence $ toPrecedence pri
                           writeIORef streamState HalfClosed
-                          let req = mkreq vh (return "")
+                          let !req = mkreq vh (return "")
                           atomically $ writeTQueue inputQ $ Input strm req
                       Nothing -> E.throwIO $ StreamError ProtocolError streamId
               Open (HasBody hdr pri) -> do
@@ -118,7 +124,7 @@ frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
                           writeIORef streamContentLength $ vhCL vh
                           readQ <- newReadBody q
                           bodySource <- mkSource readQ
-                          let req = mkreq vh (readSource bodySource)
+                          let !req = mkreq vh (readSource bodySource)
                           atomically $ writeTQueue inputQ $ Input strm req
                       Nothing -> E.throwIO $ StreamError ProtocolError streamId
               s@(Open Continued{}) -> do
@@ -147,63 +153,60 @@ frameReceiver ctx mkreq recvN = loop `E.catch` sendGoaway
                          when (isHalfClosed st) $ E.throwIO $ ConnectionError StreamClosed "header must not be sent to half closed"
                      return strm0
                  Nothing    -> do
+                     -- checkme
                      when (ftyp `notElem` [FrameHeaders,FramePriority]) $
                          E.throwIO $ ConnectionError ProtocolError "this frame is not allowed in an idel stream"
+                     csid <- readIORef currentStreamId
+                     when (streamId <= csid) $
+                         E.throwIO $ ConnectionError ProtocolError "stream identifier must not decrease"
                      when (ftyp == FrameHeaders) $ do
-                         csid <- readIORef currentStreamId
-                         if streamId <= csid then
-                             E.throwIO $ ConnectionError ProtocolError "stream identifier must not decrease"
-                           else
-                             writeIORef currentStreamId streamId
+                         writeIORef currentStreamId streamId
                          cnt <- readIORef concurrency
-                         when (cnt >= recommendedConcurrency) $ do
-                             -- Record that the stream is closed, rather than
-                             -- idle, so that receiving frames on it is only a
-                             -- stream error.
-                             consume payloadLength
-                             strm <- newStream concurrency streamId 0
-                             writeIORef (streamState strm) $ Closed $
-                                 ResetByMe $ E.toException $
-                                     StreamError RefusedStream streamId
-                             insert streamTable streamId strm
+                         when (cnt >= recommendedConcurrency) $
                              E.throwIO $ StreamError RefusedStream streamId
                      ws <- initialWindowSize <$> readIORef http2settings
-                     newstrm <- newStream concurrency streamId (fromIntegral ws)
-                     when (ftyp == FrameHeaders) $ opened newstrm
+                     newstrm <- newStream streamId (fromIntegral ws)
+                     when (ftyp == FrameHeaders) $ opened ctx newstrm
                      insert streamTable streamId newstrm
                      return newstrm
 
     consume = void . recvN
 
+initialFrame :: ByteString
+initialFrame = settingsFrame id [(SettingsMaxConcurrentStreams,recommendedConcurrency)]
+
 ----------------------------------------------------------------
 
 control :: FrameTypeId -> FrameHeader -> ByteString -> Context -> IO Bool
-control FrameSettings header@FrameHeader{flags} bs Context{http2settings, outputQ} = do
+control FrameSettings header@FrameHeader{flags} bs Context{http2settings, controlQ,firstSettings} = do
     SettingsFrame alist <- guardIt $ decodeSettingsFrame header bs
     case checkSettingsList alist of
         Just x  -> E.throwIO x
         Nothing -> return ()
     unless (testAck flags) $ do
-        modifyIORef http2settings $ \old -> updateSettings old alist
-        let frame = settingsFrame setAck []
-        enqueueControl outputQ 0 $ OSettings frame alist
+        modifyIORef' http2settings $ \old -> updateSettings old alist
+        let !frame = settingsFrame setAck []
+        sent <- readIORef firstSettings
+        let !frame' = if sent then frame else BS.append initialFrame frame
+        unless sent $ writeIORef firstSettings True
+        enqueueControl controlQ $ CSettings frame' alist
     return True
 
-control FramePing FrameHeader{flags} bs Context{outputQ} =
+control FramePing FrameHeader{flags} bs Context{controlQ} =
     if testAck flags then
         E.throwIO $ ConnectionError ProtocolError "the ack flag of this ping frame must not be set"
       else do
-        let frame = pingFrame bs
-        enqueueControl outputQ 0 $ OFrame frame
+        let !frame = pingFrame bs
+        enqueueControl controlQ $ CFrame frame
         return True
 
-control FrameGoAway _ _ Context{outputQ} = do
-    enqueueControl outputQ 0 OFinish
+control FrameGoAway _ _ Context{controlQ} = do
+    enqueueControl controlQ CFinish
     return False
 
 control FrameWindowUpdate header bs Context{connectionWindow} = do
     WindowUpdateFrame n <- guardIt $ decodeWindowUpdateFrame header bs
-    w <- (n +) <$> atomically (readTVar connectionWindow)
+    !w <- (n +) <$> atomically (readTVar connectionWindow)
     when (isWindowOverflow w) $ E.throwIO $ ConnectionError FlowControlError "control window should be less than 2^31"
     atomically $ writeTVar connectionWindow w
     return True
@@ -234,8 +237,8 @@ stream FrameHeaders header@FrameHeader{flags} bs ctx (Open JustOpened) Stream{st
         Just p  -> do
             checkPriority p streamNumber
             return p
-    let endOfStream = testEndStream flags
-        endOfHeader = testEndHeader flags
+    let !endOfStream = testEndStream flags
+        !endOfHeader = testEndHeader flags
     if endOfHeader then do
         hdr <- hpackDecodeHeader frag ctx
         return $ if endOfStream then
@@ -250,7 +253,7 @@ stream FrameHeaders header@FrameHeader{flags} bs _ (Open (Body q)) _ = do
     -- trailer is not supported.
     -- let's read and ignore it.
     HeadersFrame _ _ <- guardIt $ decodeHeadersFrame header bs
-    let endOfStream = testEndStream flags
+    let !endOfStream = testEndStream flags
     if endOfStream then do
         atomically $ writeTQueue q ""
         return HalfClosed
@@ -261,18 +264,18 @@ stream FrameHeaders header@FrameHeader{flags} bs _ (Open (Body q)) _ = do
 stream FrameData
        header@FrameHeader{flags,payloadLength,streamId}
        bs
-       Context{outputQ} s@(Open (Body q))
+       Context{controlQ} s@(Open (Body q))
        Stream{streamNumber,streamBodyLength,streamContentLength} = do
     DataFrame body <- guardIt $ decodeDataFrame header bs
-    let endOfStream = testEndStream flags
+    let !endOfStream = testEndStream flags
     len0 <- readIORef streamBodyLength
     let !len = len0 + payloadLength
     writeIORef streamBodyLength len
     when (payloadLength /= 0) $ do
-        let frame1 = windowUpdateFrame 0 payloadLength
-            frame2 = windowUpdateFrame streamNumber payloadLength
-            frame = frame1 `BS.append` frame2
-        enqueueControl outputQ 0 $ OFrame frame
+        let !frame1 = windowUpdateFrame 0 payloadLength
+            !frame2 = windowUpdateFrame streamNumber payloadLength
+            !frame = frame1 `BS.append` frame2
+        enqueueControl controlQ $ CFrame frame
     atomically $ writeTQueue q body
     if endOfStream then do
         mcl <- readIORef streamContentLength
@@ -285,16 +288,16 @@ stream FrameData
         return s
 
 stream FrameContinuation FrameHeader{flags} frag ctx (Open (Continued rfrags siz n endOfStream pri)) _ = do
-    let endOfHeader = testEndHeader flags
-        rfrags' = frag : rfrags
-        siz' = siz + BS.length frag
-        n' = n + 1
+    let !endOfHeader = testEndHeader flags
+        !rfrags' = frag : rfrags
+        !siz' = siz + BS.length frag
+        !n' = n + 1
     when (siz' > 51200) $ -- fixme: hard coding: 50K
       E.throwIO $ ConnectionError EnhanceYourCalm "Header is too big"
     when (n' > 10) $ -- fixme: hard coding
       E.throwIO $ ConnectionError EnhanceYourCalm "Header is too fragmented"
     if endOfHeader then do
-        let hdrblk = BS.concat $ reverse rfrags'
+        let !hdrblk = BS.concat $ reverse rfrags'
         hdr <- hpackDecodeHeader hdrblk ctx
         return $ if endOfStream then
                     Open (NoBody hdr pri)
@@ -305,22 +308,21 @@ stream FrameContinuation FrameHeader{flags} frag ctx (Open (Continued rfrags siz
 
 stream FrameWindowUpdate header@FrameHeader{streamId} bs _ s Stream{streamWindow} = do
     WindowUpdateFrame n <- guardIt $ decodeWindowUpdateFrame header bs
-    w <- (n +) <$> atomically (readTVar streamWindow)
+    !w <- (n +) <$> atomically (readTVar streamWindow)
     when (isWindowOverflow w) $
         E.throwIO $ StreamError FlowControlError streamId
     atomically $ writeTVar streamWindow w
     return s
 
-stream FrameRSTStream header bs _ _ strm = do
+stream FrameRSTStream header bs ctx _ strm = do
     RSTStreamFrame e <- guardIt $ decoderstStreamFrame header bs
-    let cc = Reset e
-    closed strm cc
+    let !cc = Reset e
+    closed ctx strm cc
     return $ Closed cc -- will be written to streamState again
 
 stream FramePriority header bs Context{outputQ,priorityTreeSize} s Stream{streamNumber,streamPrecedence} = do
     PriorityFrame newpri <- guardIt $ decodePriorityFrame header bs
     checkPriority newpri streamNumber
-    -- checkme: this should be tested
     oldpre <- readIORef streamPrecedence
     let !newpre = toPrecedence newpri
     writeIORef streamPrecedence newpre
@@ -332,8 +334,8 @@ stream FramePriority header bs Context{outputQ,priorityTreeSize} s Stream{stream
       else do
         mx <- delete outputQ streamNumber oldpre
         case mx of
-            Nothing -> return ()
-            Just x  -> enqueue outputQ streamNumber newpre x
+            Nothing  -> return ()
+            Just out -> enqueueOutput outputQ out
     return s
 
 -- this ordering is important
