@@ -11,6 +11,7 @@ module Network.Wai.Handler.Warp.Response (
   , addDate
   , addServer
   , hasBody
+  , replaceHeader
   ) where
 
 #ifndef MIN_VERSION_base
@@ -35,6 +36,7 @@ import Data.ByteString.Builder (byteString, Builder)
 import Data.ByteString.Builder.Extra (flush)
 import qualified Data.CaseInsensitive as CI
 import Data.Function (on)
+import Data.Hashable (hash)
 import Data.List (deleteBy)
 import Data.Maybe
 #if MIN_VERSION_base(4,5,0)
@@ -146,14 +148,14 @@ sendResponse settings conn ii req reqidxhdr src response = do
         -- and status, the response to HEAD is processed here.
         --
         -- See definition of rsp below for proper body stripping.
-        (ms, mlen) <- sendRsp conn ii ver s hs rsp
+        (ms, mlen) <- sendRsp conn ii req ver s hs rsp
         case ms of
             Nothing         -> return ()
             Just realStatus -> logger req realStatus mlen
         T.tickle th
         return ret
       else do
-        _ <- sendRsp conn ii ver s hs RspNoBody
+        _ <- sendRsp conn ii req ver s hs RspNoBody
         logger req s Nothing
         T.tickle th
         return isPersist
@@ -223,6 +225,7 @@ data Rsp = RspNoBody
 
 sendRsp :: Connection
         -> InternalInfo
+        -> Request
         -> H.HttpVersion
         -> H.Status
         -> H.ResponseHeaders
@@ -231,7 +234,7 @@ sendRsp :: Connection
 
 ----------------------------------------------------------------
 
-sendRsp conn _ ver s hs RspNoBody = do
+sendRsp conn _ _ ver s hs RspNoBody = do
     -- Not adding Content-Length.
     -- User agents treats it as Content-Length: 0.
     composeHeader ver s hs >>= connSendAll conn
@@ -239,7 +242,7 @@ sendRsp conn _ ver s hs RspNoBody = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ ver s hs (RspBuilder body needsChunked) = do
+sendRsp conn _ _ ver s hs (RspBuilder body needsChunked) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     let hdrBdy
          | needsChunked = header <> chunkedTransferEncoding body
@@ -252,7 +255,7 @@ sendRsp conn _ ver s hs (RspBuilder body needsChunked) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ ver s hs (RspStream streamingBody needsChunked th) = do
+sendRsp conn _ _ ver s hs (RspStream streamingBody needsChunked th) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     (recv, finish) <- newBlazeRecv $ reuseBufferStrategy
                     $ toBuilderBuffer (connWriteBuffer conn) (connBufferSize conn)
@@ -276,7 +279,7 @@ sendRsp conn _ ver s hs (RspStream streamingBody needsChunked th) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ _ _ (RspRaw withApp src tickle) = do
+sendRsp conn _ _ _ _ _ (RspRaw withApp src tickle) = do
     withApp recv send
     return (Nothing, Nothing)
   where
@@ -290,9 +293,10 @@ sendRsp conn _ _ _ _ (RspRaw withApp src tickle) = do
 
 -- Sophisticated WAI applications.
 -- We respect s0. s0 MUST be a proper value.
-sendRsp conn ii ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
-    sendRspFile2XX conn ii ver s0 hs path beg len isHead hook
+sendRsp conn ii req ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
+    sendRspFile2XX conn ii req ver s0 hs h path beg len isHead hook
   where
+    h = hash $ rawPathInfo req
     beg = filePartOffset part
     len = filePartByteCount part
     hs = addContentHeadersForFilePart hs0 part
@@ -301,33 +305,36 @@ sendRsp conn ii ver s0 hs0 (RspFile path (Just part) _ isHead hook) =
 
 -- Simple WAI applications.
 -- Status is ignored
-sendRsp conn ii ver _ hs0 (RspFile path Nothing idxhdr isHead hook) = do
-    efinfo <- E.try $ fileInfo ii path
+sendRsp conn ii req ver _ hs0 (RspFile path Nothing idxhdr isHead hook) = do
+    let h = hash $ rawPathInfo req
+    efinfo <- E.try $ fileInfo' ii h path
     case efinfo of
         Left (_ex :: E.IOException) ->
 #ifdef WARP_DEBUG
           print _ex >>
 #endif
-          sendRspFile404 conn ii ver hs0
+          sendRspFile404 conn ii req ver hs0
         Right finfo -> case conditionalRequest finfo hs0 idxhdr of
-          WithoutBody s         -> sendRsp conn ii ver s hs0 RspNoBody
-          WithBody s hs beg len -> sendRspFile2XX conn ii ver s hs path beg len isHead hook
+          WithoutBody s         -> sendRsp conn ii req ver s hs0 RspNoBody
+          WithBody s hs beg len -> sendRspFile2XX conn ii req ver s hs h path beg len isHead hook
 
 ----------------------------------------------------------------
 
 sendRspFile2XX :: Connection
                -> InternalInfo
+               -> Request
                -> H.HttpVersion
                -> H.Status
                -> H.ResponseHeaders
+               -> Hash
                -> FilePath
                -> Integer
                -> Integer
                -> Bool
                -> IO ()
                -> IO (Maybe H.Status, Maybe Integer)
-sendRspFile2XX conn ii ver s hs path beg len isHead hook
-  | isHead = sendRsp conn ii ver s hs RspNoBody
+sendRspFile2XX conn ii req ver s hs h path beg len isHead hook
+  | isHead = sendRsp conn ii req ver s hs RspNoBody
   | otherwise = do
       lheader <- composeHeader ver s hs
 #ifdef WINDOWS
@@ -338,7 +345,7 @@ sendRspFile2XX conn ii ver s hs path beg len isHead hook
          -- settingsFdCacheDuration is 0
          Nothing  -> return (Nothing, hook)
          Just fdc -> do
-            (fd, fresher) <- F.getFd fdc path
+            (fd, fresher) <- F.getFd' fdc h path
             return (Just fd, hook >> fresher)
       let fid = FileId path mfd
 #endif
@@ -347,10 +354,11 @@ sendRspFile2XX conn ii ver s hs path beg len isHead hook
 
 sendRspFile404 :: Connection
                -> InternalInfo
+               -> Request
                -> H.HttpVersion
                -> H.ResponseHeaders
                -> IO (Maybe H.Status, Maybe Integer)
-sendRspFile404 conn ii ver hs0 = sendRsp conn ii ver s hs (RspBuilder body True)
+sendRspFile404 conn ii req ver hs0 = sendRsp conn ii req ver s hs (RspBuilder body True)
   where
     s = H.notFound404
     hs =  replaceHeader H.hContentType "text/plain; charset=utf-8" hs0
